<?php

namespace App\Http\Controllers;

use App\Client;
use App\Utils\Utils;
use Illuminate\Http\Request;
use App\Http\Controllers\Helpers\AvatarStorageHelper as AvatarStorage;

class ClientController extends Controller
{
    public function create(Request $request)
    {
        $validator = Utils::validate($request->all(), [
            'name' => 'required|string|max:50',
            'about' => 'required|string|max:100',
            'email' => 'required|email:rfc,dns|unique:clients,email|max:50',
            'phone' => 'required|string|unique:clients,phone|max:25',
            'avatar' => 'required|array',
            'avatar.uri' => 'required|string',
            'avatar.extension' => 'required|string'
        ]);
        if ($validator->failed) {
            return $validator->error();
        }
        $avatar = new AvatarStorage($request->avatar);
        $client = Client::create([
            'name' => $request->name,
            'about' => $request->about,
            'email' => $request->email,
            'phone' => $request->phone,
            'avatar' => $avatar->url()
        ]);
        return Utils::response($client, 201);
    }

    public function paginate()
    {
        $clients = Client::orderBy('id', 'desc')->paginate(20);
        return Utils::response($clients);
    }

    public function shortlist()
    {
        $clients = Client::select(['id', 'name'])
            ->orderBy('id', 'desc')
            ->get();
        return Utils::response($clients);
    }

    public function trash()
    {
        $clients = Client::onlyTrashed()
            ->orderBy('deleted_at', 'desc')
            ->paginate(20);
        return Utils::response($clients);
    }

    public function read(Client $client)
    {
        $client->projects_count = $client->projects()->count();
        return Utils::response($client);
    }

    public function edit(Request $request, Client $client)
    {
        $validator = Utils::validate($request->all(), [
            'name' => 'required|string|max:50',
            'about' => 'required|string|max:100',
            'email' => 'required|email:rfc,dns|unique:clients,email,'.$client->id.'|max:50',
            'phone' => 'required|string|unique:clients,phone,'.$client->id.'|max:25',
            'avatar' => 'nullable|array',
            'avatar.uri' => 'required_with:avatar|string',
            'avatar.extension' => 'required_with:avatar|string'
        ]);
        if ($validator->failed) {
            return $validator->error();
        }
        $client->update([
            'name' => $request->name,
            'about' => $request->about,
            'email' => $request->email,
            'phone' => $request->phone
        ]);
        if ($request->avatar) {
            $avatar = new AvatarStorage($request->avatar);
            $client->update([ 'avatar' => $avatar->url() ]);
        }
        return Utils::response($client);
    }

    public function delete(Client $client)
    {
        $client->delete();
    }

    public function restore($id)
    {
        $client = Client::onlyTrashed()->findOrFail($id);
        $client->restore();
    }
}
