<?php

namespace App\Http\Controllers;

use App\File;
use App\Client;
use App\Utils\Utils;
use App\Mail\FileShared;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\Mail;
use App\Http\Controllers\Helpers\FileStorageHelper as FileStorage;

class FileController extends Controller
{
    public function create(Request $request)
    {
        $validator = Utils::validate($request->all(), [
            'title' => 'required|string|max:50',
            'description' => 'required|string|max:150',
            'file' => 'required|array',
            'file.uri' => 'required|string',
            'file.extension' => 'required|string|max:20'
        ]);
        if ($validator->failed) {
            return $validator->error();
        }
        $fileStorage = new FileStorage($request->file);
        $file = File::create([
            'uuid' => $fileStorage->id,
            'title' => $request->title,
            'description' => $request->description,
            'path' => $fileStorage->path(),
            'size' => $fileStorage->size(),
            'extension' => $fileStorage->extension,
            'url' => $fileStorage->url()
        ]);
        return Utils::response($file, 201);
    }

    public function paginate()
    {
        $files = File::orderBy('id', 'desc')->paginate(20);
        return Utils::response($files);
    }

    public function trash()
    {
        $files = File::onlyTrashed()
            ->orderBy('deleted_at', 'desc')
            ->paginate(20);
        return Utils::response($files);
    }

    public function read(File $file)
    {
        return Utils::response($file);
    }
    
    public function collect($uuid)
    {
        $file = File::where('uuid', $uuid)
            ->firstOrFail();
        return Utils::response($file);
    }

    public function edit(Request $request, File $file)
    {
        $validator = Utils::validate($request->all(), [
            'title' => 'required|string|max:50',
            'description' => 'required|string|max:150'
        ]);
        if ($validator->failed) {
            return $validator->error();
        }
        $file->update([
            'title' => $request->title,
            'description' => $request->description
        ]);
        return Utils::response($file);
    }

    public function delete(File $file)
    {
        $file->delete();
    }

    public function restore($id)
    {
        $file = File::onlyTrashed()->findOrFail($id);
        $file->restore();
    }

    public function share(Request $request, File $file)
    {
        $validator = Utils::validate($request->all(), [
            'client_id' => 'required|integer|exists:clients,id',
        ]);
        if ($validator->failed) {
            return $validator->error();
        }
        $client = Client::findOrFail($request->client_id);
        Mail::to($client)->send(new FileShared($file, $client));
        return Utils::response($file);
    }
}
