<?php

namespace App\Http\Controllers;

use Exception;
use App\Payment;
use App\Invoice;
use App\Utils\Utils;
use Illuminate\Http\Request;
use App\Mail\PaymentCompleted;
use Illuminate\Support\Facades\Mail;
use App\Http\Controllers\Helpers\NotificationHelper as Notification;
use App\Http\Controllers\Helpers\PaypalHelper as Paypal;
use App\Http\Controllers\Helpers\InvoiceStorageHelper as InvoiceStorage;

class PaymentController extends Controller
{
    public function create(Request $request)
    {
        $validator = Utils::validate($request->all(), [
            'uuid' => 'required|uuid',
            'pay_id' => 'required|string|unique:payments,pay_id|max:50',
            'payer_id' => 'required|string|max:50',
            'token' => 'required|string|unique:payments,token|max:50'
        ]);
        if ($validator->failed) {
            return $validator->error();
        }
        $invoice = Invoice::where('uuid', $request->uuid)->firstOrFail();
        try {
            Paypal::execute($request->pay_id, $request->payer_id);
        } catch (Exception $e) {
            return abort(404);
        }
        $invoice->payments()->create([
            'pay_id' => $request->pay_id,
            'payer_id' => $request->payer_id,
            'token' => $request->token,
            'status' => 'success'
        ]);
        $invoice->update([ 'status' => 'paid' ]);
        InvoiceStorage::store($invoice);
        Mail::to($invoice->milestone->project->client)
            ->send(new PaymentCompleted($invoice));
        $message = 'Received #'.$invoice->id;
        Notification::create('payment', $message);
    }

    public function paginate()
    {
        $payments = Payment::with('invoice.milestone.project')
            ->orderBy('id', 'desc')
            ->paginate(20);
        return Utils::response($payments);
    }

    public function cancel(Request $request)
    {
        $validator = Utils::validate($request->all(), [
            'uuid' => 'required|uuid',
            'token' => 'required|string|unique:payments,token|max:50'
        ]);
        if ($validator->failed) {
            return $validator->error();
        }
        $invoice = Invoice::where('uuid', $request->uuid)->firstOrFail();
        $invoice->payments()->create([
            'token' => $request->token,
            'status' => 'canceled'
        ]);
        $message = 'Canceled #'.$invoice->id;
        Notification::create('payment', $message);
    }
}
