<?php

namespace App\Http\Controllers;

use App\Client;
use App\Project;
use App\Utils\Utils;
use Illuminate\Http\Request;

class ProjectController extends Controller
{
    public function create(Request $request)
    {
        $validator = Utils::validate($request->all(), [
            'client_id' => 'required|integer|exists:clients,id',
            'title' => 'required|string|max:50',
            'description' => 'required|string|max:150'
        ]);
        if ($validator->failed) {
            return $validator->error();
        }
        $client = Client::findOrFail($request->client_id);
        $project = $client->projects()->create([
            'title' => $request->title,
            'description' => $request->description
        ]);
        $project->client = $project->client;
        return Utils::response($project, 201);
    }

    public function paginate()
    {
        $projects = Project::with('client')
            ->orderBy('id', 'desc')
            ->paginate(20);
        return Utils::response($projects);
    }

    public function shortlist()
    {
        $projects = Project::select([ 'projects.id', 'projects.title' ])
            ->orderBy('id', 'desc')
            ->get();
        return Utils::response($projects);
    }

    public function trash()
    {
        $projects = Project::onlyTrashed()
            ->with('client')
            ->orderBy('deleted_at', 'desc')
            ->paginate(20);
        return Utils::response($projects);
    }

    public function read(Project $project)
    {
        $project->client = $project->client;
        $project->milestones_count = $project->milestones()->count();
        $project->budget = $project->budget();
        $project->start_at = $project->startAt();
        $project->finish_at = $project->finishAt();
        return Utils::response($project);
    }

    public function edit(Request $request, Project $project)
    {
        $validator = Utils::validate($request->all(), [
            'title' => 'required|string|max:50',
            'description' => 'required|string|max:150'
        ]);
        if ($validator->failed) {
            return $validator->error();
        }
        $project->update([
            'title' => $request->title,
            'description' => $request->description
        ]);
        $project->client = $project->client;
        return Utils::response($project);
    }

    public function delete(Project $project)
    {
        $project->delete();
    }

    public function restore($id)
    {
        $project = Project::onlyTrashed()->findOrFail($id);
        $project->restore();
    }

    public function milestones(Project $project)
    {
        $milestons = $project->milestones()
            ->doesnthave('invoice')
            ->get();
        return Utils::response($milestons);
    }
}
